/* Shared library add-on to libiptables to add customized REJECT support.
 *
 * (C) 2001-2001 Harald Welte <laforge@gnumonks.org>, 
 * based on previous work (C) 2000 Jozsef Kadlecsik <kadlec@blackhole.kfki.hu>
 *
 */
#include <libiptables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_REJECT.h>

#define REJECT_PARAMS		1

#define REJECT_WITH_VALUES	7

libipt_param_values_t reject_values[REJECT_WITH_VALUES] = {
	{ IPT_ICMP_NET_UNREACHABLE, "icmp-net-unreachable",
	  "ICMP network unreachable" },
	{ IPT_ICMP_HOST_UNREACHABLE, "icmp-host-unreachable",
	  "ICMP host unreachable" },
	{ IPT_ICMP_PROT_UNREACHABLE, "icmp-proto-unreachable",
	  "ICMP protocol unreachable" },
	{ IPT_ICMP_PORT_UNREACHABLE, "icmp-port-unreachable",
	  "ICMP port unreachable (default)" },
	{ IPT_ICMP_NET_PROHIBITED, "icmp-net-prohibited",
	  "ICMP network prohibited" },
	{ IPT_ICMP_HOST_PROHIBITED, "icmp-host-prohibited",
	  "ICMP host prohibited" },
	{ IPT_TCP_RESET, "tcp-reset", "TCP RST packet" }
};
	

libipt_param_t reject_params[REJECT_PARAMS] = {
	{ { NULL, NULL },	
	  "reject-with",
	  "REJECT a packet with",
	  LIBIPT_PARAM_ENUM,
	  LIBIPT_PARAM_FLAG_NONE,
	  { values: reject_values,
	    valuenr: REJECT_WITH_VALUES },
	  { 0 },
	},
};

#if 0

/* enum needs: possible values
 * set needs: possible values, possible combinations */
 
struct reject_names {
	const char *name;
	const char *alias;
	enum ipt_reject_with with;
	const char *desc;
};

static const struct reject_names reject_table[] = {
	{"icmp-net-unreachable", "net-unreach",
		IPT_ICMP_NET_UNREACHABLE, "ICMP network unreachable"},
	{"icmp-host-unreachable", "host-unreach",
		IPT_ICMP_HOST_UNREACHABLE, "ICMP host unreachable"},
	{"icmp-proto-unreachable", "proto-unreach",
		IPT_ICMP_PROT_UNREACHABLE, "ICMP protocol unreachable"},
	{"icmp-port-unreachable", "port-unreach",
		IPT_ICMP_PORT_UNREACHABLE, "ICMP port unreachable (default)"},
#if 0
	{"echo-reply", "echoreply",
	 IPT_ICMP_ECHOREPLY, "for ICMP echo only: faked ICMP echo reply"},
#endif
	{"icmp-net-prohibited", "net-prohib",
	 IPT_ICMP_NET_PROHIBITED, "ICMP network prohibited"},
	{"icmp-host-prohibited", "host-prohib",
	 IPT_ICMP_HOST_PROHIBITED, "ICMP host prohibited"},
	{"tcp-reset", "tcp-reset",
	 IPT_TCP_RESET, "TCP RST packet"}
};

static void
print_reject_types()
{
	unsigned int i;

	printf("Valid reject types:\n");

	for (i = 0; i < sizeof(reject_table)/sizeof(struct reject_names); i++) {
		printf("    %-25s\t%s\n", reject_table[i].name, reject_table[i].desc);
		printf("    %-25s\talias\n", reject_table[i].alias);
	}
	printf("\n");
}

/* Saves the union ipt_targinfo in parsable form to stdout. */

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"REJECT options:\n"
"--reject-with type              drop input packet and send back\n"
"                                a reply packet according to type:\n");

	print_reject_types();
}

static struct option opts[] = {
	{ "reject-with", 1, 0, '1' },
	{ 0 }
};

/* Allocate and initialize the target. */
static void
init(struct ipt_entry_target *t, unsigned int *nfcache)
{
	struct ipt_reject_info *reject = (struct ipt_reject_info *)t->data;

	/* default */
	reject->with = IPT_ICMP_PORT_UNREACHABLE;

	/* Can't cache this */
	*nfcache |= NFC_UNKNOWN;
}

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      struct ipt_entry_target **target)
{
	struct ipt_reject_info *reject = (struct ipt_reject_info *)(*target)->data;
	unsigned int limit = sizeof(reject_table)/sizeof(struct reject_names);
	unsigned int i;

	switch(c) {
	case '1':
		if (check_inverse(optarg, &invert))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --reject-with");
		for (i = 0; i < limit; i++) {
			if ((strncasecmp(reject_table[i].name, optarg, strlen(optarg)) == 0)
			    || (strncasecmp(reject_table[i].alias, optarg, strlen(optarg)) == 0)) {
				reject->with = reject_table[i].with;
				return 1;
			}
		}
		/* This due to be dropped late in 2.4 pre-release cycle --RR */
		if (strncasecmp("echo-reply", optarg, strlen(optarg)) == 0
		    || strncasecmp("echoreply", optarg, strlen(optarg)) == 0)
			fprintf(stderr, "--reject-with echo-reply no longer"
				" supported\n");
		exit_error(PARAMETER_PROBLEM, "unknown reject type `%s'",optarg);
	default:
		/* Fall through */
	}
	return 0;
}

/* Final check; nothing. */
static void final_check(unsigned int flags)
{
}

/* Prints out ipt_reject_info. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_target *target,
      int numeric)
{
	const struct ipt_reject_info *reject
		= (const struct ipt_reject_info *)target->data;
	unsigned int i;

	for (i = 0; i < sizeof(reject_table)/sizeof(struct reject_names); i++) {
		if (reject_table[i].with == reject->with)
			break;
	}
	printf("reject-with %s ", reject_table[i].name);
}

/* Saves ipt_reject in parsable form to stdout. */
static void save(const struct ipt_ip *ip, const struct ipt_entry_target *target)
{
	const struct ipt_reject_info *reject
		= (const struct ipt_reject_info *)target->data;
	unsigned int i;

	for (i = 0; i < sizeof(reject_table)/sizeof(struct reject_names); i++)
		if (reject_table[i].with == reject->with)
			break;

	printf("--reject-with %s ", reject_table[i].name);
}
#endif

static int init(libipt_plugin_t *pl, void *data)
{
	struct ipt_reject_info *rejinfo = (struct ipt_reject_info *)data;

	rejinfo->with = IPT_ICMP_PORT_UNREACHABLE;

	return TRUE;
}

static int struct2rule(libipt_plugin_t *pl, void *data)
{
	libipt_param_t *pm;
	struct ipt_reject_info *rejinfo = (struct ipt_reject_info *)data;

	pm = list_entry(pl->params.next, libipt_param_t, list);
	if (libipt_param_set(pm) == TRUE)
		rejinfo->with = pm->value.set;

	return TRUE;
}

static int check(libipt_plugin_t *pl, void *data)
{
	/* FIXME */
}

static int rule2struct()
{ 
	/* FIXME */
}

static libipt_plugin_t reject = {
	{ NULL, NULL },
	LIBIPT_PLUGIN_TARGET,
	NULL,
	"REJECT",
	"REJECT Target",
	LIBIPTABLES_VERSION,
	IPT_ALIGN(sizeof(struct ipt_reject_info)),
	IPT_ALIGN(sizeof(struct ipt_reject_info)),
	{ NULL, NULL },
	&init,
	&struct2rule,
	&check,
	&rule2struct
};

void _init(void)
{
	libipt_prepare_params(&reject.params, reject_params, REJECT_PARAMS);
	libipt_register_plugin(&reject);
}
