/* libpkttables.c, Version $Revision: 701 $
 *
 * high level API for iptables rule operations, part of the 
 * netfilter/iptables infrastructure of linux 2.4 / 2.5
 *
 * (C) 2001-2002 by Harald Welte <laforge@gnumonks.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 
 *  as published by the Free Software Foundation
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  
 * $Id: libpkttables.c 701 2004-10-07 14:35:28Z laforge $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <dlfcn.h>
#include <libpkttables.h>

#define LIBPKTT_RULEBUF_SIZE 1024

#define LIBPKTT_PLUGIN_PREFIX "libpktt_"

#ifndef LIBPKTT_PLUGIN_DIR
#define LIBPKTT_PLUGIN_DIR "/usr/local/lib/libpkttables"
#endif

#ifndef PROC_SYS_MODPROBE
#define PROC_SYS_MODPROBE "/proc/sys/kernel/modprobe"
#endif

/***********************************************************************
 * PRIVATE FUNCTIONS
 ***********************************************************************/

static struct list_head libpktt_plugins = LIST_HEAD_INIT(libpktt_plugins);
static struct list_head libpktt_handles = LIST_HEAD_INIT(libpktt_handles);

/* return 1 for a valid plugin filename */
static int valid_pluginname(const struct dirent *de)
{
	if (!strncmp(LIBPKTT_PLUGIN_PREFIX, de->d_name, 
		     strlen(LIBPKTT_PLUGIN_PREFIX)))
		return 1;
	
	/* DON'T USE TRUE/FALSE HERE !!! */
	return 0;
}

/* get struct libpktt_plugin for given name */
static struct libpktt_plugin *plugin_find(char *plname)
{
	struct list_head *tmp;
	struct libpktt_plugin *pl;

	list_for_each(tmp, &libpktt_plugins) {
		pl = list_entry(tmp, struct libpktt_plugin, list);
		DEBUGP("pl=%p pl->name='%s'\n", pl, pl->name);
		if (!strcmp(pl->name, plname))
			return pl;
	}
	/* why not ?
	return list_named_find(&libpktt_plugins, plname);
	*/
	return NULL;
}

static struct libpktt_param *param_find(char *pmname) 
{
	struct list_head *tmp, *tmp2;
	struct libpktt_plugin *pl;
	struct libpktt_param *pm;
	
	list_for_each(tmp, &libpktt_plugins) {
		pl = list_entry(tmp, struct libpktt_plugin, list);
		list_for_each(tmp2, &pl->params) {
			pm = list_entry(tmp2, struct libpktt_param, list);
			DEBUGP("plname: %s, param: %s\n", pl->name, pm->name);
			if (!strcmp(pm->name, pmname))
				return pm;
		}
	}

	return NULL;
}

/* verify that parameters are unique */
static int check_paramnames(struct libpktt_param *params)
{
	struct list_head *tmp;
	struct libpktt_param *pm;

	list_for_each(tmp, &params->list) {
		pm = list_entry(tmp, struct libpktt_param, list);
		if (param_find(pm->name)) {
			DEBUGP("parameter collision: %s\n", pm->name);
			return FALSE;
		}
	}

	return TRUE;
}

/*********************************************************************** 
 * EXPORTED API FOR PLUGINS
 ***********************************************************************/

/* libpktt_prepare_params -
 *
 */
int libpktt_prepare_params(struct list_head *list,
			  struct libpktt_param *params,
			  int param_count) {
	int i;
	INIT_LIST_HEAD(list);

	for (i = 0; i < param_count; i++) {
		list_add(&params[i].list, list);
	}

	DEBUGP("returning from libpktt_prepare_params\n");
	return TRUE;
}

/* libpktt_register_plugin - register a plugin with the library core
 * @pl: plugin to be registered
 *
 * This function is to be called by the init() function of any plugin
 * registering with the core library.
 *
 * return value: TRUE if operation suceeds, FALSE in case of any error.
 */
int libpktt_register_plugin(libpktt_plugin_t *pl) {

	if (!pl)
		return FALSE;

	if (plugin_find(pl->name))
		return FALSE;

	if (check_paramnames(list_entry(&pl->params, struct libpktt_param,
					list)) != TRUE)
		return FALSE;
	
	DEBUGP("registering plugin %s\n", pl->name);
	list_add(&pl->list, &libpktt_plugins);

	return TRUE;
}

/*********************************************************************** 
 * EXPORTED API FOR APPLICATIONS
 ***********************************************************************/

/* libpktt_init - initialize the library 
 *
 * This functions needs to be called prior any use. It initializes the library
 * and loads all available plugins.
 *
 * A program would usually call libpktt_init() once, and later call 
 * libpktt_get_handle() for every context.
 *
 * return value: TRUE if library initialization succeeds, FALSE in case of any
 * error (OOM, no plugins found, error while loading plugin, ...)
 */
int libpktt_init()
{
	struct dirent **namelist;
	int n;
	void *ptr;
	char path[sizeof(LIBPKTT_PLUGIN_DIR) + strlen(LIBPKTT_PLUGIN_PREFIX)
		  + LIBPKTT_NAMELEN + strlen(".so") + 3];

	/* get list of plugins */
	n = scandir(LIBPKTT_PLUGIN_DIR, &namelist, valid_pluginname, alphasort);
	if (n < 0) {
		perror("scandir");
		goto free_plugins;
	}

	while (n--) {
		sprintf(path, LIBPKTT_PLUGIN_DIR "/%s", namelist[n]->d_name);
		DEBUGP("loading plugin %s\n", namelist[n]->d_name);
		if ((ptr = dlopen(path, RTLD_NOW))) {
			/* Found plugin (and is initialized) */

			struct libpktt_plugin *pl;
			char *pluginname;
			char *tmpstr;

			tmpstr = strdup(namelist[n]->d_name);

			if (!tmpstr) {
				DEBUGP("ENOMEM while allocating tmpstr\n");
				free(namelist[n]);
				goto free_plugins;
			}

			/* we only want XXX out of libpktt_XXX.so */
			pluginname = tmpstr + strlen(LIBPKTT_PLUGIN_PREFIX);
			*(pluginname+strlen(pluginname)-3) = '\0';

			pl = plugin_find(pluginname);
			if (!pl) {
				DEBUGP("ERROR: no plugin %s after opening %s\n",
					pluginname, namelist[n]->d_name);
				free(pluginname);
				free(namelist[n]);
				goto free_plugins;
			} 

			DEBUGP("found plugin %s after opening %s\n", 
				pluginname, namelist[n]->d_name);
			/* set address of plugin */
			pl->addr = ptr;

		} else {
			DEBUGP("ERROR dlopening %s: %s\n", path, dlerror());
		}
		
		free(namelist[n]);
	}

	return TRUE;

free_plugins:
	/* FIXME: unload plugins */

	return FALSE;
}

/* libpktt_exit - destruct library, unload all plugins 
 *
 * This function is the de-initialization function of the library.
 * 
 * Return value: TRUE if de-initialization successful, FALSE if there are
 * still open library handles.
 *
 */
int libpktt_exit()
{
	struct list_head *tmp, *tmp2;
	libpktt_plugin_t *pl;

	/* FIXME: checkif any handles in use */
	if (!list_empty(&libpktt_handles))
		return FALSE;

	list_for_each_safe(tmp, tmp2, &libpktt_plugins) {
		list_del(tmp);
		pl = list_entry(tmp, libpktt_plugin_t, list);
		dlclose(pl->addr);
	}

	return TRUE;
}

/* libpktt_get_handle - create a new library handle 
 *
 * A library handle is necessarry for any subsequent function call of
 * libpkttables.  All application-specific context is associated with this
 * handle.
 *
 * It is safe to use multiple library handles in case of multiple rule
 * operation contexts.
 *
 * return value: the newly created library handle, NULL in case of any error
 */
libpktt_handle_t *libpktt_get_handle() {
	struct libpktt_handle *h;

	h = (libpktt_handle_t *) malloc(sizeof(*h));
	if (!h)
		return NULL;
	memset(h, 0, sizeof(*h));

	INIT_LIST_HEAD(&h->tables);
	INIT_LIST_HEAD(&h->list);

	list_add_tail(&h->list, &libpktt_handles);

	return h;
}

void libpktt_rule_free(libpktt_rule_t *rule) {
	/* FIXME */
	free(rule);
}

/* libpktt_chain_free - free (delete) a chain
 * @chain: chain to be freed.
 *
 * A call to this function frees the chain structure and all associated rules,
 * if any
 */
void libpktt_chain_free(libpktt_chain_t *chain) {
	struct list_head *tmp, *tmp2;
	libpktt_rule_t *rule;

	list_for_each_safe(tmp, tmp2, &chain->rules) {
		list_del(tmp);
		rule = list_entry(tmp, libpktt_rule_t, list);
		libpktt_rule_free(rule);
	}
	free(chain);
}

void libpktt_table_free(libpktt_table_t *table) {
	struct list_head *tmp, *tmp2;
	libpktt_chain_t *chain;

	list_for_each_safe(tmp, tmp2, &table->chains) {
		list_del(tmp);
		chain = list_entry(tmp, libpktt_chain_t, list);
		libpktt_chain_free(chain);
	}
}

/* libpktt_free_handle - destroy a library handle 
 * h: Handle to be freed
 *
 * This function frees a library handle.  All associated  data structures like
 * tables, chains, rules, ... are free()d, too.
 */
int libpktt_free_handle(libpktt_handle_t *h) {

	/* FIXME: clean up all tables/chains/rules/... */
	struct list_head *tmp, *tmp2;

	list_for_each_safe(tmp, tmp2, &h->tables) {
		libpktt_table_t *table;

		list_del(tmp);
		table = list_entry(tmp, libpktt_table_t, list);
		libpktt_table_free(table);
	}

	list_del(&h->list);
	free(h);
	return TRUE;
}

/* libpktt_first_plugin - iterate over list of plugins */
libpktt_plugin_t *libpktt_first_plugin(libpktt_handle_t *h)
{
	libpktt_plugin_t *pl;

	pl = list_entry(&libpktt_plugins.next, libpktt_plugin_t, list);
	return pl;
}

/* libpktt_next_plugin - iterate over list of plugins */
libpktt_plugin_t *libpktt_next_plugin(libpktt_handle_t *h, libpktt_plugin_t *prev)
{
	libpktt_plugin_t *pl;

	if (prev->list.next == &libpktt_plugins)
		return NULL;

	pl = list_entry(prev->list.next, libpktt_plugin_t, list);
	return pl;
}

/* libpktt_get_pluginfo -  get more information about a plugin
 * plname: name of plugin
 *
 * This function resolves the plugin information structure for the plugin of
 * the given 'plname' name.
 *
 * Return value: plugin data structure
 */
libpktt_plugin_t *libpktt_get_pluginfo(char *plname) {
	struct libpktt_plugin *pl;

	pl = plugin_find(plname);
	if (!pl)
		return NULL;

	return pl;
}


/* libpktt_rule_create - create a new rule handle 
 * @chain: chain in which this  rule is created
 *
 * Create a new rule in the given chain
 *
 */
libpktt_rule_t *
libpktt_rule_create(libpktt_chain_t *chain)
{
	libpktt_rule_t *ruleh;

	ruleh = (libpktt_rule_t *) malloc(sizeof(*ruleh));
	if (!ruleh)
		return NULL;
	memset(ruleh, 0, sizeof(*ruleh));

	ruleh->chain = chain;
	INIT_LIST_HEAD(&ruleh->list);

	ruleh->rulebuf_len = LIBPKTT_RULEBUF_SIZE;
	ruleh->rulebuf_offset = 0;
	ruleh->rulebuf = malloc(ruleh->rulebuf_len);
	if (!ruleh->rulebuf) {
		free(ruleh);
		return NULL;
	}

	return ruleh;
}

static int create_val_list(struct list_head *vallist,
			   libpktt_relem_t *relem)
{
	struct list_head *tmp;
	libpktt_param_t *param;
	libpktt_param_value_t *vale;

	list_for_each(tmp, &relem->params) {
		param = list_entry(tmp, libpktt_param_t, list);
		vale = (libpktt_param_value_t *) malloc(sizeof(*vale));
		if (!vale) {
			/* FIXME: free already alloc()ed vallist entries */
			return 0;
		}
		memset(vale, 0, sizeof(*vale));
		vale->pm = param;
		list_add(&vale->list, vallist);
	}

	return 1;
}

/* create a new relem handle */
libpktt_relem_handle_t *
libpktt_relem_create(libpktt_rule_t *ruleh, libpktt_relem_t *relem)
{
	libpktt_relem_handle_t *relemh;

	/* allocate handle */
	relemh = (libpktt_relem_handle_t *) malloc(sizeof(*relemh));
	if (!relemh)
		goto err_relemh;
	memset(relemh, 0, sizeof(*relemh));

	/* allocate buffer for kernel structure */
	relemh->kernelstruct = malloc(relem->size);
	if (!relemh->kernelstruct)
		goto err_struct;
	memset(relemh->kernelstruct, 0, relem->size);

	relemh->rule = ruleh;
	relemh->relem = relem;

	/* allocate+link value_list */
	INIT_LIST_HEAD(&relemh->val_list);
	if (!create_val_list(&relemh->val_list, relem))
		goto err_vallist;

	return relemh;

err_vallist:
	free(relemh->kernelstruct);
err_struct:
	free(relemh);
err_relemh:
	return NULL;
}

libpktt_param_value_t *get_value_by_param(libpktt_relem_handle_t *relemh, 
					libpktt_param_t *pm)
{
	struct list_head *tmp;
	libpktt_param_value_t *val;

	list_for_each(tmp, &relemh->val_list) {
		val = list_entry(tmp, libpktt_param_value_t, list);
		if (val->pm == pm)
			return val;
	}
	return NULL;
}

/* create relem out of caller-supplied values list */
int libpktt_struct2relem(struct libpktt_plugin *pl,
			struct list_head *values,
			struct libpktt_rule *rule)
{
	libpktt_param_value_t *val;

	list_for_each_entry(tmp, values, list) {


	}
}

/* add relem to specified rule */
int libpktt_relem2rule(libpktt_relem_handle_t *relemh, libpktt_rule_t *ruleh)
{


}



/***********************************************************************
 * CHAIN / TABLE HANDLING
 ***********************************************************************/

/* libpktt_table_create - create a handle for a given table
 * @h: library handle
 * @name: name of the table
 * @protocol: protocol number (0800 for ipv4, ...)
 *
 * return value: pointer libpktt_table_t, NULL in case of any error.
 */
libpktt_table_t *libpktt_table_create(libpktt_handle_t *h, char *name,
				    unsigned int protocol)
{
	libpktt_table_t *table;

	table = (libpktt_table_t *) malloc(sizeof(*table));
	if (!table)
		return NULL;
	memset(table, 0, sizeof(*table));
	INIT_LIST_HEAD(&table->list);
	INIT_LIST_HEAD(&table->chains);

	table->h = h;
	strncpy(table->name, name, sizeof(table->name)-1);
	table->protocol = protocol;
	list_add_tail(&table->list, &h->tables);

	return table;
}

/* libpktt_table_attach_chain - attach chain to hook of table
 * @table: table to which chain is to be attached
 * @chain: chain which is to be attached
 * @hooknum: hook number to which chain is to be attached
 *
 * return value: TRUE if attach_chain succeeds, FALSE in case of any error.
 */
int libpktt_table_attach_chain(libpktt_table_t *table,
			      libpktt_chain_t *chain,
			      int hooknum)
{
	if (chain->table != table)
		return FALSE;

	if (!(hooknum & table->valid_hooks))
		return FALSE;

	table->builtin_chains[hooknum] = chain;
	return TRUE;
}

/* libpktt_chain_create - create a new chain in a given table
 * @table: table in which chain is created
 * @name: name of table
 *
 * Creates a new chain called 'name' in the given table.
 *
 * Return value: Chain handle if operation succeeds, FALSE otherwise
 */
libpktt_chain_t *libpktt_chain_create(libpktt_table_t *table, const char *name)
{
	libpktt_chain_t *chain;

	/* FIXME: check if chain with 'name' already exists in table */

	chain = (libpktt_chain_t *) malloc(sizeof(*chain));
	if (!chain)
		return NULL;
	memset(chain, 0, sizeof(*chain));
	INIT_LIST_HEAD(&chain->list);
	INIT_LIST_HEAD(&chain->rules);

	strncpy(chain->name, name, sizeof(chain->name)-1);
	chain->table = table;
	list_add_tail(&chain->list, &table->chains);
	
	return chain;
}

/* libpktt_chain_append - append a rule to a chain 
 * @chain: chain to which rule is added
 * @rule: the rule that is added
 * */
int libpktt_chain_append(libpktt_chain_t *chain, libpktt_rule_t *rule)
{
	list_add_tail(&rule->list, &chain->rules);
	return TRUE;
}


