/* libpkttables.h, Version $Revision: 698 $
 *
 * high level API for iptables rule operations, part of the 
 * netfilter/iptables infrastructure of linux 2.4 / 2.5
 *
 * (C) 2001-2003 by Harald Welte <laforge@netfilter.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 
 *  as published by the Free Software Foundation
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  
 * $Id: libpkttables.h 698 2004-01-09 00:33:02Z laforge $
 */
#ifndef _LIBPKTTABLES_H
#define _LIBPKTTABLES_H

#include <netinet/in.h>
#include "listhelp.h"		/* which includes linuxlist.h */


#ifndef LIBPKTTABLES_VERSION
#define LIBPKTTABLES_VERSION	"2.0.0"
#endif

#define LIBPKTT_MAX_HOOKS		8
#define LIBPKTT_NAMELEN			32
#define LIBPKTT_DESC_LEN		1024	/* maximum description length */

#if 1
#include <stdio.h>
#define DEBUGP(format, args...) fprintf(stderr, format, ## args)
#endif

#ifndef TRUE
#define TRUE 	0
#endif
#ifndef FALSE
#define FALSE 	1
#endif

/* ugly forward declarations */
typedef struct libpktt_plugin libpktt_plugin_t;
typedef struct libpktt_plugin libpktt_relem_t;
typedef struct libpktt_chain libpktt_chain_t;
typedef struct libpktt_rule libpktt_rule_t;
typedef struct libpktt_table libpktt_table_t;

/* needed for LIBPKTT_PARAM_ENUM / SET / BITMASK */
typedef struct libpktt_param_values {
	unsigned int value;
	const char *string;
	const char *description;
} libpktt_param_values_t;

/* numeric values start with 0x0, others with 0x1 */
#define LIBPKTT_PARAM_T_MASK	0xFF00
#define LIBPKTT_PARAM_T_NUMERIC  0x0000
	#define LIBPKTT_PARAM_T_SIGNED	0x0000
	#define LIBPKTT_PARAM_T_UNSIGNED 0x0100
#define LIBPKTT_PARAM_T_SPECIAL	0x1000
#define LIBPKTT_PARAM_T_ADDINFO	0x2000

enum libpktt_param_type {
	LIBPKTT_PARAM_NONE = 0x0000,

	LIBPKTT_PARAM_INT8 = LIBPKTT_PARAM_T_SIGNED + 1,
	LIBPKTT_PARAM_INT16,
	LIBPKTT_PARAM_INT32,
	LIBPKTT_PARAM_INT64,

	LIBPKTT_PARAM_UINT8 = LIBPKTT_PARAM_INT8 | LIBPKTT_PARAM_T_UNSIGNED,
	LIBPKTT_PARAM_UINT16,
	LIBPKTT_PARAM_UINT32,
	LIBPKTT_PARAM_UINT64,

	LIBPKTT_PARAM_BOOL = LIBPKTT_PARAM_T_SPECIAL + 1,
	LIBPKTT_PARAM_IPADDR,
	LIBPKTT_PARAM_IP6ADDR,
	LIBPKTT_PARAM_PORT,
	LIBPKTT_PARAM_RANGE,
	LIBPKTT_PARAM_STRING,

	LIBPKTT_PARAM_ENUM = LIBPKTT_PARAM_T_ADDINFO + 1,
	LIBPKTT_PARAM_SET,
	LIBPKTT_PARAM_BITMASK,

	LIBPKTT_PARAM_CALLBACK = 0x4000,

};

#define LIBPKTT_PARAM_FLAG_NONE		0x0000
#define LIBPKTT_PARAM_FLAG_MANDATORY	0x0001
#define LIBPKTT_PARAM_FLAG_INVERTIBLE	0x0002
#define LIBPKTT_PARAM_FLAG_SET		0x1000	/* option set (in use) */
#define LIBPKTT_PARAM_FLAG_INVERTED	0x2000	/* option inverted */

/* description of a parameter, as specified by plugin */
typedef struct libpktt_param {
	struct list_head list;		/* linked list within plugin */
	struct libpktt_plugin *pl;	/* to which plugin we belong */

#define LIBPKTT_MAX_NAMELEN 32
	char name[LIBPKTT_MAX_NAMELEN];

	char description[LIBPKTT_DESC_LEN];

	/* type of the returned value (LIBPKTT_PARAM_...) */
	enum libpktt_param_type type;

	/* flags (mandatory, optional, ...) */
	u_int32_t flags;

	struct {
		libpktt_param_values_t *values;
		unsigned int valuenr;
		char **possibilities;
	} type_addinfo;


	/* length of this parameters' value, if applicable */
	u_int32_t len;
	
} libpktt_param_t;

typedef struct libpktt_param_value {
	struct list_head list; 		/* linked list between values */
	struct libpktt_param *pm;	/* for which param are we? */
	u_int32_t flags;		/* PARAM_SET, ... */
	union {
		u_int8_t	b;
		u_int8_t	ui8;
		u_int16_t	ui16;
		u_int32_t	ui32;
		u_int64_t	ui64;
		int8_t		i8;
		int16_t		i16;
		int32_t		i32;
		int64_t		i64;
		void		*ptr;
		char		*string;
		u_int32_t	set;
		struct {
			u_int32_t	ipaddr;
			u_int32_t	mask;
		} ip;
		struct {
			struct in6_addr	ip6addr;
			struct in6_addr mask;
		} ip6;
	} value;			/* value of the parameter */
} libpktt_param_value_t;


enum libpktt_plugin_type {
	LIBPKTT_PLUGIN_NONE = 0,
	LIBPKTT_PLUGIN_MATCH,
	LIBPKTT_PLUGIN_TARGET,
};


/* libpktt_plugin_t, the structure registered by a match extension
 * to describe the match */

struct libpktt_plugin {
	struct list_head list;		/* linked list of all matches */
	void *addr;			/* address of plugin */
	
//	u_int16_t protocol;		/* ipv4/ipv6/generic/... */

	enum libpktt_plugin_type type;	/* type of plugin */

	char name[LIBPKTT_NAMELEN];		/* name */
	char description[LIBPKTT_DESC_LEN]; /* description */
	
	const char *version;

	/* Size of match data */
	size_t size;

	/* Size of match data relevant for userspace comparison purposes */
	size_t userspacesize;

	/* linked list of match parameters, NULL if none */
	struct list_head params;
	
	/* Initialize the match */
	int (*init)(struct libpktt_plugin *pl, void *data); /* FIXME: nfcache*/

	/* Function which parses (interprets) the descriptive libpkttables
	 * structures into the kernel readable structure */
	int (*struct2rule)(struct libpktt_plugin *pl,
			   struct list_head *values,
			   struct libpktt_rule *rule);

	/* Function to do a final check on the rule */
	int (*check)(struct libpktt_plugin *pl, 
		     struct list_head *values,
		     struct libpktt_rule *rule);
	
	/* Function to convert an in-kernel rule into libpkttables structures */
	int (*rule2struct)();

};

typedef struct libpktt_handle {
	struct list_head list;		/* list of handles */
	struct list_head tables;	/* list of tables in this handle */
} libpktt_handle_t;

struct libpktt_rule {
	struct list_head list;		/* linked list of rules in chain */
	libpktt_chain_t *chain;		/* our master chain handle */

	libpktt_handle_t *ipth;		/* our master ipthandle, if orphan */

	void *rulebuf;			/* pointer to  rulebuffer */
	int rulebuf_len;		/* current size of rulebuf */
	int rulebuf_offset;		/* current offset in rulebuf */
};

struct libpktt_chain {
	struct list_head list;		/* list of chains in table */
	libpktt_table_t *table;		/* our master table */

	char name[LIBPKTT_NAMELEN];	/* name of the chain */
	struct list_head rules;		/* list of rules in chain */
};

struct libpktt_table {
	struct list_head list;		/* list of tables in handle */
	libpktt_handle_t	*h;	/* handle */

	char name[LIBPKTT_NAMELEN];	/* name of the table */
	struct list_head chains;	/* list of chains in this table */
	unsigned int valid_hooks;	/* bitmask of valid hooks */
	unsigned int protocol;		/* l3 protocol */
	libpktt_chain_t *builtin_chains[LIBPKTT_MAX_HOOKS];	/* builtin chains */
};


libpktt_rule_t 
libpktt_create_rule(libpktt_handle_t *ipth, unsigned int protocol);


typedef struct libpktt_relem_handle {
	libpktt_rule_t *rule;		/* our master rulehandle */
	libpktt_relem_t *relem;		/* the relem we belong to */

	struct list_head val_list;	/* the list of values for relem */
	void *kernelstruct;
} libpktt_relem_handle_t;

libpktt_relem_handle_t
libpktt_gethandle_relem(libpktt_rule_t *ruleh, libpktt_relem_t *relem);




/* EXPORTED API TO PLUGINS */

/* fill in all the list_head members */
int libpktt_prepare_params(struct list_head *list,
			  struct libpktt_param *params,
			  int param_count);

/* register a plugin with the library */
int libpktt_register_plugin(libpktt_plugin_t *pl);

#define libpktt_param_value_set(param)	((param->flags & LIBPKTT_PARAM_FLAG_SET) ? TRUE : FALSE)
#define libpktt_param_value_inverted(param) ((param->flags & LIBPKTT_PARAM_FLAG_INVERTED) ? TRUE : FALSE)
#define libpktt_param_value_unsigned(param) (((type & LIBPKTT_PARAM_T_MASK) == LIBPKTT_PARAM_T_UNSIGNED) ? TRUE : FALSE)

#define libpktt_param_set_value(value) (value->flags |= LIBPKTT_PARAM_FLAG_SET)
#define libpktt_param_invert_value(value) (value->flags |= LIBPKTT_PARAM_FLAG_INVERTED)

/* EXPORTED API TO PROGRAMS */

int libpktt_init(void);
int libpktt_exit(void);

libpktt_handle_t *libpktt_get_handle(void);
int libpktt_free_handle(libpktt_handle_t *h);

/* iterate over all available plugins */
libpktt_plugin_t *libpktt_first_plugin(libpktt_handle_t *h);
libpktt_plugin_t *libpktt_next_plugin(libpktt_handle_t *h, libpktt_plugin_t *prev);

/* get information about a plugin */
libpktt_plugin_t *libpktt_get_pluginfo(char *pname);

/* create a new table */
libpktt_table_t *libpktt_table_create(libpktt_handle_t *h, char *name,
					unsigned int protocol);;

/* attach a chain to a given hook at a given table */
int libpktt_table_attach_chain(libpktt_table_t *table, libpktt_chain_t *chain,
			      int hooknum);

/* create a new chain */
libpktt_chain_t *libpktt_chain_create(libpktt_table_t *table, const char *name);

/* append a rule to a chain */
int libpktt_chain_append(libpktt_chain_t *chain, libpktt_rule_t *rule);

libpktt_rule_t *libpktt_rule_create(libpktt_chain_t *chain);

#endif /* LIBPKTTABLES_H */

#include <libpkttables_print.h>
