/* libpkttables_pares.c, Version $Revision: 698 $
 *
 * high level API for iptables rule operations, part of the 
 * netfilter/iptables infrastructure of linux 2.4 / 2.5
 *
 * (C) 2001-2003 by Harald Welte <laforge@netfilter.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 
 *  as published by the Free Software Foundation
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  
 * $Id: libpkttables_parse.c 698 2004-01-09 00:33:02Z laforge $
 */

#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <netinet/in.h>
#include <netdb.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <libpkttables.h>


/* PORT NUMBERS */

static int
service_to_port(const char *name, const char *proto)
{
	struct servent *service;

	if ((service = getservbyname(name, proto)) != NULL)
		return ntohs((unsigned short) service->s_port);

		return -1;
}

#if 0
static void
parse_portrange(const char *portstring, const char *proto, u_int16_t *ports)
{
	char *buffer;
	char *cp;

	buffer = strdup(portstring);
	if ((cp = strchr(buffer, ':')) == NULL)
		ports[0] = ports[1] = parse_udp_port(buffer);
	else {
		*cp = '\0';
		cp++;

		ports[0] = buffer[0] ? parse_udp_port(buffer) : 0;
		ports[1] = cp[0] ? parse_udp_port(cp) : 0xFFFF;
	}
	free(buffer);
}

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_udp *udpinfo = (struct ipt_udp *)(*match)->data;

	switch (c) {
	case '1':
		if (*flags & UDP_SRC_PORTS)
			exit_error(PARAMETER_PROBLEM,
				   "Only one `--source-port' allowed");
		if (check_inverse(optarg, &invert))
			optind++;
		parse_udp_ports(argv[optind-1], udpinfo->spts);
		if (invert)
			udpinfo->invflags |= IPT_UDP_INV_SRCPT;
		*flags |= UDP_SRC_PORTS;
		*nfcache |= NFC_IP_SRC_PT;
		break;

	case '2':
		if (*flags & UDP_DST_PORTS)
			exit_error(PARAMETER_PROBLEM,
				   "Only one `--destination-port' allowed");
		if (check_inverse(optarg, &invert))
			optind++;
		parse_udp_ports(argv[optind-1], udpinfo->dpts);
		if (invert)
			udpinfo->invflags |= IPT_UDP_INV_DSTPT;
		*flags |= UDP_DST_PORTS;
		*nfcache |= NFC_IP_DST_PT;
		break;

	default:
		return 0;
	}

	return 1;
}

/* Final check; we don't care. */
static void
final_check(unsigned int flags)
{
}

static char *
port_to_service(int port)
{
	struct servent *service;

	if ((service = getservbyport(htons(port), "udp")))
		return service->s_name;

	return NULL;
}

static void
print_port(u_int16_t port, int numeric)
{
	char *service;

	if (numeric || (service = port_to_service(port)) == NULL)
		printf("%u", port);
	else
		printf("%s", service);
}

static void
print_ports(const char *name, u_int16_t min, u_int16_t max,
	    int invert, int numeric)
{
	const char *inv = invert ? "!" : "";

	if (min != 0 || max != 0xFFFF || invert) {
		printf("%s", name);
		if (min == max) {
			printf(":%s", inv);
			print_port(min, numeric);
		} else {
			printf("s:%s", inv);
			print_port(min, numeric);
			printf(":");
			print_port(max, numeric);
		}
		printf(" ");
	}
}

/* Prints out the union ipt_matchinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_match *match, int numeric)
{
	const struct ipt_udp *udp = (struct ipt_udp *)match->data;

	printf("udp ");
	print_ports("spt", udp->spts[0], udp->spts[1],
		    udp->invflags & IPT_UDP_INV_SRCPT,
		    numeric);
	print_ports("dpt", udp->dpts[0], udp->dpts[1],
		    udp->invflags & IPT_UDP_INV_DSTPT,
		    numeric);
	if (udp->invflags & ~IPT_UDP_INV_MASK)
		printf("Unknown invflags: 0x%X ",
		       udp->invflags & ~IPT_UDP_INV_MASK);
}

/* Saves the union ipt_matchinfo in parsable form to stdout. */
static void save(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
	const struct ipt_udp *udpinfo = (struct ipt_udp *)match->data;

	if (udpinfo->spts[0] != 0
	    || udpinfo->spts[1] != 0xFFFF) {
		if (udpinfo->invflags & IPT_UDP_INV_SRCPT)
			printf("! ");
		if (udpinfo->spts[0]
		    != udpinfo->spts[1])
			printf("--sport %u:%u ",
			       udpinfo->spts[0],
			       udpinfo->spts[1]);
		else
			printf("--sport %u ",
			       udpinfo->spts[0]);
	}

	if (udpinfo->dpts[0] != 0
	    || udpinfo->dpts[1] != 0xFFFF) {
		if (udpinfo->invflags & IPT_UDP_INV_DSTPT)
			printf("! ");
		if (udpinfo->dpts[0]
		    != udpinfo->dpts[1])
			printf("--dport %u:%u ",
			       udpinfo->dpts[0],
			       udpinfo->dpts[1]);
		else
			printf("--dport %u ",
			       udpinfo->dpts[0]);
	}
}

#endif

static int string_to_number(const char *s, long *ret, long min, long max)
			    
{
	long number;
	char *end;

	/* handle hex, octal, etc. */
	errno = 0;
	number = strtol(s, &end, 0);
	if (*end == '\0' && end != s) {
		/* we parsed a number, let's see if we want this */
		if (errno != ERANGE && min <= number && number <= max) {
			*ret = number;
			return TRUE;
		}
	}
	return FALSE;
}

static int string_to_unumber(const char *s, unsigned long *ret, 
			     unsigned long min, unsigned long max)
			    
{
	unsigned long number;
	char *end;

	/* handle hex, octal, etc. */
	errno = 0;
	number = strtoul(s, &end, 0);
	if (*end == '\0' && end != s) {
		/* we parsed a number, let's see if we want this */
		if (errno != ERANGE && min <= number && number <= max) {
			*ret = number;
			return TRUE;
		}
	}
	return FALSE;
}

static struct in_addr *
dotted_to_addr(const char *dotted)
{
	static struct in_addr addr;
	unsigned char *addrp;
	char *p, *q;
	long onebyte;
	int i;
	char buf[20];

	/* copy dotted string, because we need to modify it */
	strncpy(buf, dotted, sizeof(buf) - 1);
	addrp = (unsigned char *) &(addr.s_addr);

	p = buf;
	for (i = 0; i < 3; i++) {
		if ((q = strchr(p, '.')) == NULL)
			return (struct in_addr *) NULL;

		*q = '\0';
		if (string_to_number(p, &onebyte, 0, 255) == -1)
			return (struct in_addr *) NULL;

		addrp[i] = (unsigned char) onebyte;
		p = q + 1;
	}

	/* we've checked 3 bytes, now we check the last one */
	if (string_to_number(p, &onebyte, 0, 255) == -1)
		return (struct in_addr *) NULL;

	addrp[3] = (unsigned char) onebyte;

	return &addr;
}

static struct in_addr *
network_to_addr(const char *name)
{
	struct netent *net;
	static struct in_addr addr;

	if ((net = getnetbyname(name)) != NULL) {
		if (net->n_addrtype != AF_INET)
			return (struct in_addr *) NULL;
		addr.s_addr = htonl((unsigned long) net->n_net);
		return &addr;
	}

	return (struct in_addr *) NULL;
}

static void
inaddrcpy(struct in_addr *dst, struct in_addr *src)
{
	/* memcpy(dst, src, sizeof(struct in_addr)); */
	dst->s_addr = src->s_addr;
}

static struct in_addr *
host_to_addr(const char *name, unsigned int *naddr)
{
	struct hostent *host;
	struct in_addr *addr;
	unsigned int i;

	*naddr = 0;
	if ((host = gethostbyname(name)) != NULL) {
		if (host->h_addrtype != AF_INET ||
		    host->h_length != sizeof(struct in_addr))
			return (struct in_addr *) NULL;

		while (host->h_addr_list[*naddr] != (char *) NULL)
			(*naddr)++;
		addr = calloc(*naddr, sizeof(struct in_addr));
		for (i = 0; i < *naddr; i++)
			inaddrcpy(&(addr[i]),
				  (struct in_addr *) host->h_addr_list[i]);
		return addr;
	}

	return (struct in_addr *) NULL;
}


static struct in_addr *
parse_hostnetwork(const char *name, unsigned int *naddrs)
{
	struct in_addr *addrp, *addrptmp;

	if ((addrptmp = dotted_to_addr(name)) != NULL ||
	    (addrptmp = network_to_addr(name)) != NULL) {
		addrp = malloc(sizeof(struct in_addr));
		inaddrcpy(addrp, addrptmp);
		*naddrs = 1;
		return addrp;
	}
	if ((addrp = host_to_addr(name, naddrs)) != NULL)
		return addrp;

	//exit_error(PARAMETER_PROBLEM, "host/network `%s' not found", name);
	return NULL;
}

static struct in_addr *
parse_mask(char *mask)
{
	static struct in_addr maskaddr;
	struct in_addr *addrp;
	long bits;

	if (mask == NULL) {
		/* no mask at all defaults to 32 bits */
		maskaddr.s_addr = 0xFFFFFFFF;
		return &maskaddr;
	}
	if ((addrp = dotted_to_addr(mask)) != NULL)
		/* dotted_to_addr already returns a network byte order addr */
		return addrp;
	if (string_to_number(mask, &bits, 0, 32) == -1)
		//exit_error(PARAMETER_PROBLEM, "invalid mask `%s' specified", mask);
		return NULL;
	if (bits != 0) {
		maskaddr.s_addr = htonl(0xFFFFFFFF << (32 - bits));
		return &maskaddr;
	}

	maskaddr.s_addr = 0L;
	return &maskaddr;
}

static void
parse_hostnetworkmask(const char *name, struct in_addr **addrpp,
		      struct in_addr *maskp, unsigned int *naddrs)
{
	struct in_addr *addrp;
	char buf[256];
	char *p;
	int i, j, k, n;

	strncpy(buf, name, sizeof(buf) - 1);
	if ((p = strrchr(buf, '/')) != NULL) {
		*p = '\0';
		addrp = parse_mask(p + 1);
	} else
		addrp = parse_mask(NULL);
	inaddrcpy(maskp, addrp);

	/* if a null mask is given, the name is ignored, like in "any/0" */
	if (maskp->s_addr == 0L)
		strcpy(buf, "0.0.0.0");

	addrp = *addrpp = parse_hostnetwork(buf, naddrs);
	n = *naddrs;
	for (i = 0, j = 0; i < n; i++) {
		addrp[j++].s_addr &= maskp->s_addr;
		for (k = 0; k < j - 1; k++) {
			if (addrp[k].s_addr == addrp[j - 1].s_addr) {
				(*naddrs)--;
				j--;
				break;
			}
		}
	}
}

/* parse routines for different supported types */

static int parse_number(char *cmd, libpktt_param_value_t *v)
{
	unsigned long umin, umax;
	long min, max;
	int ret;
	long dummy;

	switch (v->pm->type) {
		case LIBPKTT_PARAM_INT8:
			min = SCHAR_MIN;
			max = SCHAR_MAX;
			ret = string_to_number(cmd, &dummy, min, max);
			v->value.i8 = dummy;
			break;
		case LIBPKTT_PARAM_UINT8:
			umin = 0;
			umax = UCHAR_MAX;
			ret = string_to_unumber(cmd, &dummy, umin, umax);
			v->value.ui8 = dummy;
			break;
		case LIBPKTT_PARAM_INT16:
			min = SHRT_MIN;
			max = SHRT_MAX;
			ret = string_to_number(cmd, &dummy, min, max);
			v->value.i16 = dummy;
			break;
		case LIBPKTT_PARAM_UINT16:
			umin = 0;
			umax = USHRT_MAX;
			ret = string_to_unumber(cmd, &dummy, umin, umax);
			v->value.ui16 = dummy;
			break;
		case LIBPKTT_PARAM_INT32:	
			min = INT_MIN;
			max = INT_MAX;
			ret = string_to_number(cmd, &dummy, min, max);
			v->value.i32 = dummy;
			break;
		case LIBPKTT_PARAM_UINT32:
			umin = 0;
			umax = ULONG_MAX;
			ret = string_to_unumber(cmd, &dummy, umin, umax);
			v->value.ui32 = dummy;
			break;
		case LIBPKTT_PARAM_INT64:
		case LIBPKTT_PARAM_UINT64:
		default:
			DEBUGP("unsupported numeric type\n");
			ret = FALSE;
			break;
	}

	return ret;
}


static int parse_ipaddr(const char *cmd, libpktt_param_value_t *v)
{
	int ret;

	ret = inet_pton(AF_INET, cmd, &v->value.ip.ipaddr);
	if (ret > 0)
		return TRUE;

	return FALSE;
}

static int parse_ip6addr(const char *cmd, libpktt_param_value_t *v)
{
	int ret;

	ret = inet_pton(AF_INET6, cmd, &v->value.ip6.ip6addr);
	if (ret > 0)
		return TRUE;

	return FALSE;
}

static int parse_string(const char *cmd, libpktt_param_value_t *v)
{
	if (strlen(cmd) > v->pm->len)
		return FALSE;

	v->value.string = strdup(cmd);
	if (!v->value.string)
		return FALSE;

	return TRUE;
}

static int parse_port(const char *cmd, libpktt_param_value_t *v)
{
	long portnum;

	if (string_to_number(cmd, &portnum, 0, 65535) != -1 ||
	    (portnum = service_to_port(cmd, NULL)) != -1) {

		v->value.ui16 = portnum;
		return TRUE;
	}
		
	return FALSE;
}

static int get_valnum_by_name(libpktt_param_values_t *values, int num, 
			      const char *name)
{
	int i;
	for (i = 0; i < num; i++) {
		if (!strcmp(name, values[i].string)) {
			return values[i].value;
		}
	}
	return -1;
}

static int parse_enum(const char *cmd, libpktt_param_value_t *v)
{
	int i = get_valnum_by_name(v->pm->type_addinfo.values, 
			v->pm->type_addinfo.valuenr, cmd);

	if (i >= 0)
		return TRUE;

	return FALSE;
}

static int parse_bitmask(const char *cmd, libpktt_param_value_t *v)
{
	char *cmdcp = strdup(cmd);
	char *curcmd;
	u_int32_t retmask = 0;

	if (!cmdcp)
		return FALSE;

	curcmd = strtok(cmdcp, ",| ");
	do {
		int i = get_valnum_by_name(v->pm->type_addinfo.values,
					   v->pm->type_addinfo.valuenr, curcmd);

		if (i < 0) {
			free(cmdcp);
			return FALSE;
		}

		retmask |= (1 << i);

	} while ((curcmd = strtok(NULL, ",| ")));
	

	free(cmdcp);
	return TRUE;
}

static int parse_bool(const char *cmd, libpktt_param_value_t *v)
{
	v->value.b = 1;
	return TRUE;
}

/***********************************************************************
 * MAIN PARSER ROUTINE
 ***********************************************************************/

static int parse_param(char *cmd, libpktt_param_value_t *v)
{
	enum libpktt_param_type type = v->pm->type;
	int ret = FALSE;

	/* FIXME: deal with '!' inversion */

	switch (type & LIBPKTT_PARAM_T_MASK) {
	case LIBPKTT_PARAM_T_NUMERIC:
		ret = parse_number(cmd, v);
		break;
	case LIBPKTT_PARAM_T_SPECIAL:
		switch (type) {
		case LIBPKTT_PARAM_STRING:
			ret = parse_string(cmd, v);
			break;
		case LIBPKTT_PARAM_RANGE:
			//FIXME: ret = parse_range(cmd, v);
			break;
		case LIBPKTT_PARAM_PORT:
			ret = parse_port(cmd, v);
			break;
		case LIBPKTT_PARAM_IP6ADDR:
			ret = parse_ip6addr(cmd, v);
			break;
		case LIBPKTT_PARAM_IPADDR:
			ret = parse_ipaddr(cmd, v);
			break;
		case LIBPKTT_PARAM_BOOL:
			ret = parse_bool(cmd, v);
			break;
		default:
			ret = FALSE;
			break;
		}
		break;
	case LIBPKTT_PARAM_T_ADDINFO:
		switch (type) {
		case LIBPKTT_PARAM_ENUM:
			ret = parse_enum(cmd, v);
		case LIBPKTT_PARAM_SET:
			/* FIXME */
			break;
		case LIBPKTT_PARAM_BITMASK:
			ret = parse_bitmask(cmd, v);
			break;
		default:
			ret = FALSE;
			break;
		}
		break;

	case LIBPKTT_PARAM_CALLBACK:
		/* FIXME */
		break;
	default:
		ret = FALSE;
		break;
	}

	if (ret == TRUE)
		v->flags |= LIBPKTT_PARAM_FLAG_SET;

	return ret;
}

int libpkttables_parse_plugin_cmd(char *command, size_t len, 
				 libpktt_plugin_t *pl)
{
	
	return 0;
}

