/* libpkttables_print.c, Version $Revision: 697 $
 *
 * high level API for iptables rule operations, part of the 
 * netfilter/iptables infrastructure of linux 2.4 / 2.5
 *
 * (C) 2001-2003 by Harald Welte <laforge@netfilter.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 
 *  as published by the Free Software Foundation
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  
 * $Id: libpkttables_print.c 697 2003-07-15 15:28:17Z laforge $
 */

#include <libpkttables.h>
#include <libpkttables_print.h>

/* Internal string datatype and support functions */

typedef struct _string {
	char *string;	/* begining of string */
	char *pos;	/* current write position */
	size_t size;	/* total available size of string */
} libpktt_string_t;

#define string_init(_str, _chr, _len) \
	do { _str->string = _str->pos = _chr; _str->size = _len; } while(0);
#define string_space(_str) ((_str->string + _str->size) - _str->pos)
#define printstr(_str, _fmt, _args...) \
	_str->pos += snprintf(_str->pos, string_space(_str), _fmt, ## _args);

/* internal functions */
static int param_unsigned(enum libpktt_param_type type)
{
	if ((type & LIBPKTT_PARAM_T_MASK) == LIBPKTT_PARAM_T_UNSIGNED)
		return TRUE;

	return FALSE;
}

static int print_paramtype(libpktt_string_t *outstr, 
			   enum libpktt_param_type type)
{
	char *oldpos = outstr->pos;

	/* if numeric */
	if ((type & 0xF000) == LIBPKTT_PARAM_T_NUMERIC) {
		if (param_unsigned(type) == TRUE)
			printstr(outstr, "unsigned ");

		switch (type & ~LIBPKTT_PARAM_T_UNSIGNED) {
			case LIBPKTT_PARAM_INT8:
				printstr(outstr,"8bit int");
				break;
			case LIBPKTT_PARAM_INT16:
				printstr(outstr, "16bit int");
				break;
			case LIBPKTT_PARAM_INT32:
				printstr(outstr, "32bit int");
				break;
			case LIBPKTT_PARAM_INT64:
				printstr(outstr, "64bit int");
				break;
			default:
				printstr(outstr, "unknown int type");
				break;
		}
	} else {
		switch(type) {
			case LIBPKTT_PARAM_STRING:
				printstr(outstr, "string");
				break;
			case LIBPKTT_PARAM_BOOL:
				printstr(outstr, "boolean");
				break;
			case LIBPKTT_PARAM_IPADDR:
				printstr(outstr, "IP address");
				break;
			case LIBPKTT_PARAM_IP6ADDR:
				printstr(outstr, "IPv6 address");
				break;
			case LIBPKTT_PARAM_ENUM:
				printstr(outstr, "enumeration");
				break;
			case LIBPKTT_PARAM_BITMASK:
				prinstr(outstr, "bitmask");
				break;
/*
			case LIBPKTT_PARAM_SET:
				break;
*/
			default:
				printstr(outstr, "unknown type");
		}
	}

	return (outstr->pos - oldpos);
}

static int print_param_addinfo(libpktt_string_t *outstr, libpktt_param_t *pm)
{
	unsigned int i;
	libpktt_param_values_t *values;
	char *oldpos = outstr->pos;

	if ((pm->type & 0xF000) != LIBPKTT_PARAM_T_ADDINFO)
		return 0;

	switch (pm->type) {
		case LIBPKTT_PARAM_BITMASK:
		case LIBPKTT_PARAM_SET:
			/* FIXME: how to print possible combinations ? */
		case LIBPKTT_PARAM_ENUM:
			values = pm->type_addinfo.values;
			for (i = 0; i < pm->type_addinfo.valuenr; i++) {
				printstr(outstr, "                 %-25s %s\n", 
					 values[i].string,
					 values[i].description);
			}
			break;
		default:
			printstr(outstr, "unknown addinfo!!!");
			break;
	}
	
	return (outstr->pos - oldpos);
}

static int print_paraminfo(libpktt_string_t *outstr, libpktt_param_t *pm)
{
	char *oldpos = outstr->pos;

	/* print name */
	printstr(outstr, " %-15s", pm->name);
	/* description */
	printstr(outstr, "%s", pm->description);
	/* and type */
	printstr(outstr, " (");
	print_paramtype(outstr, pm->type);
	printstr(outstr, ")\n");
	print_param_addinfo(outstr, pm);

	return (outstr->pos - oldpos);
}

/* EXPORTED FUNCTIONS */

int libpktt_print_paraminfo(char *output, size_t len, libpktt_param_t *pm)
{
	libpktt_string_t _outstr, *outstr = &_outstr;

	string_init(outstr, output, len);

	if (print_paraminfo(outstr, pm))
		return TRUE;

	return FALSE;
}

int libpktt_print_pluginhelp(char *output, size_t len, libpktt_plugin_t *pl)
{
	struct list_head *tmp;
	libpktt_param_t *pm;
	libpktt_string_t _outstr, *outstr = &_outstr;

	string_init(outstr, output, len);

	switch (pl->type) {
		case LIBPKTT_PLUGIN_MATCH:
			printstr(outstr, "Match ");
			break;
		case LIBPKTT_PLUGIN_TARGET:
			printstr(outstr, "Target ");
			break;
		default:
			printstr(outstr, "unknown plugin ");
			break;
	}

	printstr(outstr, "%s options:\n", pl->name);

	list_for_each(tmp, &pl->params) {
		pm = list_entry(tmp, libpktt_param_t, list);
		print_paraminfo(outstr, pm);
	}
	
	return TRUE;
}

